/*
Copyright (C) 2006 - 2009 Evan Teran
                   eteran@alum.rit.edu

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#ifndef SYMBOLMANAGER_20060814_H_
#define SYMBOLMANAGER_20060814_H_

#include "Types.h"
#include "API.h"

#include <QList>
#include <QHash>
#include <QMap>
#include <QSet>
#include <QSharedPointer>
#include <QString>

class EDB_EXPORT SymbolManager {
public:
	SymbolManager() {}

public:
	struct Symbol {
		QString			file;
		QString			name;
		QString			name_no_prefix;
		edb::address_t	address;
		quint32			size;
		char			type;
		
		bool is_code() const { return type == 't' || type == 'T' || type == 'P'; }
		bool is_data() const { return !is_code(); }
		bool is_weak() const { return type == 'W'; }
	};
	
	typedef QSharedPointer<Symbol> SymbolP;
	
public:
	const QList<SymbolP> symbols() const;
	const SymbolP find(const QString &name) const;
	const SymbolP find(edb::address_t address) const;
	const SymbolP find_near_symbol(edb::address_t address) const;
	void clear();
	void load_symbol_file(const QString &filename, edb::address_t base);
	void load_symbols(const QString &symbol_directory);

private:
	bool process_symbol_file(const QString &f, edb::address_t base, const QString &library_filename);
	void add_symbol(const SymbolP &sym);
	
private:
	QString							symbol_directory_;
	QSet<QString>					symbol_files_;
	QList<SymbolP>					symbols_;
	QMap<edb::address_t, SymbolP>	symbols_by_address_;
	QHash<QString, SymbolP>			symbols_by_name_;
};

#endif

