/*
Copyright (C) 2006 - 2009 Evan Teran
                   eteran@alum.rit.edu

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#ifndef ANALYZER_20080630_H_
#define ANALYZER_20080630_H_

#include "DebuggerPluginInterface.h"
#include "AnalyzerInterface.h"
#include "Types.h"
#include "MemRegion.h"
#include <QSet>
#include <QMap>

class QMenu;

class Analyzer : public QObject, public AnalyzerInterface, public DebuggerPluginInterface {
	Q_OBJECT
	Q_INTERFACES(DebuggerPluginInterface)
	Q_CLASSINFO("author", "Evan Teran")
	Q_CLASSINFO("url", "http://www.codef00.com")

public:
	Analyzer();
	
public:
	virtual QMenu *menu(QWidget *parent = 0);
	virtual QList<QAction *> cpu_context_menu();
	
private:
	virtual void private_init();
	virtual QWidget *options_page();
	
public:
	virtual void analyze(const MemRegion &region);
	virtual FunctionMap functions(const MemRegion &region) const;
	virtual AddressCatagory catagory(edb::address_t address) const;
	virtual void invalidate_analysis(const MemRegion &region);
	virtual void invalidate_analysis();

private:
	void bonus_stack_frames_helper(Function &info) const;
	void bonus_stack_frames(FunctionMap &results);
	void bonus_symbols(const MemRegion &region, FunctionMap &results);
	void bonus_entry_point(const MemRegion &region, FunctionMap &results) const;
	void bonus_main(const MemRegion &region, FunctionMap &results) const;
	
private:
	QByteArray md5_region(const MemRegion &region) const;
	bool find_containing_function(edb::address_t address, Function &function) const;
	bool is_inside_known(const MemRegion &region, edb::address_t address);
	bool is_stack_frame(edb::address_t address) const;
	bool is_thunk(edb::address_t address) const;
	edb::address_t module_entry_point(const MemRegion &region) const;
	int walk_all_functions(FunctionMap &results, const MemRegion &region, QSet<edb::address_t> &walked_functions);
	void find_calls_from_known(const MemRegion &region, FunctionMap &results, QSet<edb::address_t> &walked_functions);
	void find_function_calls(const MemRegion &region, FunctionMap &results);
	void find_function_end(Function &function, edb::address_t end_address, QSet<edb::address_t> &found_functions, const FunctionMap &results);
	void fix_overlaps(FunctionMap &function_map);
	void set_function_types(FunctionMap &results);
	void set_function_types_helper(Function &info) const;
	void update_results_entry(FunctionMap &results, edb::address_t address) const;

Q_SIGNALS:
	void update_progress(int);

public Q_SLOTS:
	void do_ip_analysis();
	void do_view_analysis();
	
private:
	void do_analysis(const MemRegion &region); 
	
private:
	struct RegionInfo {
		FunctionMap	analysis;
		QByteArray	md5;
		bool		fuzzy;
	};
	
	QMenu						*menu_;
	QMap<MemRegion, RegionInfo>	analysis_info_;
};

#endif
