/*
Copyright (C) 2006 - 2009 Evan Teran
                   eteran@alum.rit.edu

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "DebuggerCoreBase.h"
#include "X86Breakpoint.h"

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
DebuggerCoreBase::DebuggerCoreBase() : active_thread_(0), pid_(0) {
}

//------------------------------------------------------------------------------
// Name: ~DebuggerCoreBase()
// Desc: destructor
//------------------------------------------------------------------------------
DebuggerCoreBase::~DebuggerCoreBase() {
}

//------------------------------------------------------------------------------
// Name: clear_breakpoints()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerCoreBase::clear_breakpoints() {
	if(attached()) {	
		breakpoints_.clear();
	}
}

//------------------------------------------------------------------------------
// Name: add_breakpoint(edb::address_t address)
// Desc: 
//------------------------------------------------------------------------------
QSharedPointer<Breakpoint> DebuggerCoreBase::add_breakpoint(edb::address_t address) {


	if(!find_breakpoint(address)) {
		QSharedPointer<Breakpoint> bp(new X86Breakpoint(address));
		breakpoints_[address] = bp;
		return bp;
	}
	
	return QSharedPointer<Breakpoint>();
}

//------------------------------------------------------------------------------
// Name: find_breakpoint(edb::address_t address) const
// Desc: 
//------------------------------------------------------------------------------
QSharedPointer<Breakpoint> DebuggerCoreBase::find_breakpoint(edb::address_t address) {
	if(attached()) {
		BreakpointState::iterator it = breakpoints_.find(address);
		if(it != breakpoints_.end()) {
			return it.value();
		}
	}
	return QSharedPointer<Breakpoint>();
}


//------------------------------------------------------------------------------
// Name: remove_breakpoint(edb::address_t address)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerCoreBase::remove_breakpoint(edb::address_t address) {

	// TODO: assert paused
	if(attached()) {
		BreakpointState::iterator it = breakpoints_.find(address);
		if(it != breakpoints_.end()) {
			breakpoints_.erase(it);
		}
	}
}

//------------------------------------------------------------------------------
// Name: backup_breakpoints() const
// Desc: 
//------------------------------------------------------------------------------
DebuggerCoreBase::BreakpointState DebuggerCoreBase::backup_breakpoints() const {
	return breakpoints_;
}

//------------------------------------------------------------------------------
// Name: open(const QString &path, const QString &cwd, const QStringList &args)
// Desc: 
//------------------------------------------------------------------------------
bool DebuggerCoreBase::open(const QString &path, const QString &cwd, const QStringList &args) {
	return open(path, cwd, args, QString());
}

//------------------------------------------------------------------------------
// Name: pid() const
// Desc: returns the pid of the currently debugged process
//------------------------------------------------------------------------------
edb::pid_t DebuggerCoreBase::pid() const {
	return pid_;
}

//------------------------------------------------------------------------------
// Name: attached() const
// Desc: 
//------------------------------------------------------------------------------
bool DebuggerCoreBase::attached() const {
	return pid() != 0;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
QList<edb::tid_t> DebuggerCoreBase::thread_ids() const {
	return QList<edb::tid_t>();
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
edb::tid_t DebuggerCoreBase::active_thread() const {
	return active_thread_;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerCoreBase::set_active_thread(edb::tid_t) {
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
int DebuggerCoreBase::breakpoint_size() const {
	return X86Breakpoint::size;
}
