/*
Copyright (C) 2006 - 2009 Evan Teran
                   eteran@alum.rit.edu

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "DialogOptions.h"
#include "Debugger.h"

#include <QFontDialog>
#include <QFileDialog>
#include <QFont>
#include <QCloseEvent>
#include <QToolBox>

#include "ui_dialog_options.h"

//------------------------------------------------------------------------------
// Name: DialogOptions(QWidget *parent)
// Desc: 
//------------------------------------------------------------------------------
DialogOptions::DialogOptions(QWidget *parent) : QDialog(parent), ui(new Ui::DialogOptions), toolbox_(0) {
	ui->setupUi(this);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
DialogOptions::~DialogOptions() {
	delete ui;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
QString DialogOptions::fontFromDialog(const QString &default_font) {
	QFont oldFont;
	oldFont.fromString(default_font);
	return QFontDialog::getFont(0, oldFont, this).toString();
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DialogOptions::addOptionsPage(QWidget *page) {
	
	if(toolbox_ == 0) {
		delete ui->tabWidget->findChild<QLabel *>("label_plugins");
		QWidget *const tab			= ui->tabWidget->findChild<QLabel *>("tab_plugins");
		QGridLayout *const layout	= ui->tabWidget->findChild<QGridLayout *>("tab_plugins_layout");
		toolbox_ = new QToolBox(tab);
		layout->addWidget(toolbox_, 0, 0, 1, 1);
	}
	
	toolbox_->addItem(page, QIcon(":/debugger/images/edb32-preferences-plugin.png"), page->windowTitle());
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
QString DialogOptions::directoryFromDialog() {
	return QFileDialog::getExistingDirectory(
    	this,
    	tr("Choose a directory"),
    	QString(),
    	QFileDialog::ShowDirsOnly | QFileDialog::DontResolveSymlinks);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DialogOptions::on_btnTTY_clicked() {
	const QString filename = QFileDialog::getOpenFileName(
		this, 
		tr("Choose Your Terminal Program"));
		
	ui->txtTTY->setText(filename);
}

//------------------------------------------------------------------------------
// Name: on_btnStackFont_clicked()
// Desc: 
//------------------------------------------------------------------------------
void DialogOptions::on_btnStackFont_clicked() {
	ui->txtStackFont->setText(fontFromDialog(ui->txtStackFont->text()));
}

//------------------------------------------------------------------------------
// Name: on_btnMemoryFont_clicked()
// Desc: 
//------------------------------------------------------------------------------
void DialogOptions::on_btnMemoryFont_clicked() {
	ui->txtMemoryFont->setText(fontFromDialog(ui->txtMemoryFont->text()));
}

//------------------------------------------------------------------------------
// Name: on_btnRegisterFont_clicked()
// Desc: 
//------------------------------------------------------------------------------
void DialogOptions::on_btnRegisterFont_clicked() {
	ui->txtRegisterFont->setText(fontFromDialog(ui->txtRegisterFont->text()));
}

//------------------------------------------------------------------------------
// Name: on_btnDisFont_clicked()
// Desc: 
//------------------------------------------------------------------------------
void DialogOptions::on_btnDisFont_clicked() {
	ui->txtDisFont->setText(fontFromDialog(ui->txtDisFont->text()));
}

//------------------------------------------------------------------------------
// Name: on_btnSymbolDir_clicked()
// Desc: 
//------------------------------------------------------------------------------
void DialogOptions::on_btnSymbolDir_clicked() {
	const QString s = directoryFromDialog();
				
	if(!s.isEmpty()) {
		ui->txtSymbolDir->setText(s);
	}
}

//------------------------------------------------------------------------------
// Name: on_btnSessionDir_clicked()
// Desc: 
//------------------------------------------------------------------------------
void DialogOptions::on_btnSessionDir_clicked() {
	const QString s = directoryFromDialog();
				
	if(!s.isEmpty()) {
		ui->txtSessionDir->setText(s);
	}
}

//------------------------------------------------------------------------------
// Name: on_btnPluginDir_clicked()
// Desc: 
//------------------------------------------------------------------------------
void DialogOptions::on_btnPluginDir_clicked() {
	const QString s = directoryFromDialog();
	
	if(!s.isEmpty()) {
		ui->txtPluginDir->setText(s);
	}
}

//------------------------------------------------------------------------------
// Name: showEvent(QShowEvent *event)
// Desc: 
//------------------------------------------------------------------------------
void DialogOptions::showEvent(QShowEvent *event) {
	QDialog::showEvent(event);
	
	ui->rdoSytntaxATT->setChecked(syntax == ATT);
	ui->rdoSytntaxIntel->setChecked(syntax != ATT);

	ui->rdoDetach->setChecked(close_behavior == Detach);
	ui->rdoKill->setChecked(close_behavior != Detach);

	ui->rdoBPEntry->setChecked(initial_breakpoint == EntryPoint);
	ui->rdoBPMain->setChecked(initial_breakpoint != EntryPoint);
	
	ui->chkTTY->setChecked(tty_enabled);
	ui->txtTTY->setText(tty_command);
	
	ui->chkZerosAreFilling->setChecked(zeros_are_filling);
	ui->chkUppercase->setChecked(uppercase_disassembly);
	
	ui->chkFindMain->setChecked(find_main);
	ui->chkWarnDataBreakpoint->setChecked(warn_on_no_exec_bp);
	
	ui->spnMinString->setValue(min_string_length);
	
	ui->txtStackFont->setText(stack_font);
	ui->txtMemoryFont->setText(data_font);
	ui->txtRegisterFont->setText(registers_font);
	ui->txtDisFont->setText(disassembly_font);

	ui->txtSymbolDir->setText(symbol_path);
	ui->txtPluginDir->setText(plugin_path);
	ui->txtSessionDir->setText(session_path);
	
	ui->chkDataShowAddress->setChecked(data_show_address);
	ui->chkDataShowHex->setChecked(data_show_hex);
	ui->chkDataShowAscii->setChecked(data_show_ascii);
	ui->chkDataShowComments->setChecked(data_show_comments);
	ui->cmbDataWordWidth->setCurrentIndex(WidthToIndex(data_word_width));
	ui->cmbDataRowWidth->setCurrentIndex(WidthToIndex(data_row_width));
	
	ui->chkAddressSemicolon->setChecked(show_address_separator);
}

//------------------------------------------------------------------------------
// Name: WidthToIndex(int n)
// Desc: 
//------------------------------------------------------------------------------
int DialogOptions::WidthToIndex(int n) {
	switch(n) {
	case 16:
		return 4;
	case 8:
		return 3;
	case 4:
		return 2;
	case 2:
		return 1;
	default:
		return 0;
	}
}

//------------------------------------------------------------------------------
// Name: closeEvent(QCloseEvent *event)
// Desc: 
//------------------------------------------------------------------------------
void DialogOptions::closeEvent(QCloseEvent *event) {

	if(ui->rdoSytntaxIntel->isChecked()) {
		syntax = Intel;
	} else if(ui->rdoSytntaxATT->isChecked()) {
		syntax = ATT;
	}
	
	if(ui->rdoDetach->isChecked()) {
		close_behavior = Detach;
	} else if(ui->rdoKill->isChecked()) {
		close_behavior = Terminate;
	}
	
	stack_font				= ui->txtStackFont->text();
	data_font				= ui->txtMemoryFont->text();	
	registers_font			= ui->txtRegisterFont->text();
	disassembly_font		= ui->txtDisFont->text();
	tty_command				= ui->txtTTY->text();
	tty_enabled				= ui->chkTTY->isChecked();
	zeros_are_filling		= ui->chkZerosAreFilling->isChecked();
	uppercase_disassembly	= ui->chkUppercase->isChecked();
	
	symbol_path				= ui->txtSymbolDir->text();
	plugin_path				= ui->txtPluginDir->text();
	session_path			= ui->txtSessionDir->text();
	
	if(ui->rdoBPMain->isChecked()) {
		initial_breakpoint = MainSymbol;
	} else if(ui->rdoBPEntry->isChecked()) {
		initial_breakpoint = EntryPoint;
	}
	
	warn_on_no_exec_bp		= ui->chkWarnDataBreakpoint->isChecked();
	find_main				= ui->chkFindMain->isChecked();
	
	show_address_separator	= ui->chkAddressSemicolon->isChecked();

	min_string_length		= ui->spnMinString->value();
	
	data_show_address		= ui->chkDataShowAddress->isChecked();
	data_show_hex			= ui->chkDataShowHex->isChecked();
	data_show_ascii			= ui->chkDataShowAscii->isChecked();
	data_show_comments		= ui->chkDataShowComments->isChecked();
	data_word_width			= 1 << ui->cmbDataWordWidth->currentIndex();
	data_row_width			= 1 << ui->cmbDataRowWidth->currentIndex();
	
	event->accept();
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DialogOptions::accept() {
	close();
}
